package ru.dbotthepony.mc.otm.block.entity;

import com.google.common.collect.ImmutableList;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.*;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.PatternStorageMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.atomic.AtomicInteger;

public class BlockEntityPatternStorage extends BlockEntityMattery implements IMatterGridCell, IPatternStorage {
	public BlockEntityPatternStorage(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.PATTERN_STORAGE, p_155229_, p_155230_);
	}

	public final MatteryContainer patterns = new MatteryContainer(this::setChanged, 3 * 3) {
		@Override
		public void setChanged(int slot, ItemStack new_state, ItemStack old_state) {
			if (grid != null && !ItemStack.isSameItemSameTags(new_state, old_state)) {
				if (!old_state.isEmpty()) {
					old_state.getCapability(MatteryCapability.PATTERN).ifPresent(cap -> cap.getStoredPatterns().forEach(grid::onPatternRemoved));
				}

				if (!new_state.isEmpty()) {
					new_state.getCapability(MatteryCapability.PATTERN).ifPresent(cap -> cap.getStoredPatterns().forEach(grid::onPatternAdded));
				}
			}

			super.setChanged(slot, new_state, old_state);
		}

		@Override
		public int getMaxStackSize() {
			return 1;
		}
	};

	private final LazyOptional<IItemHandler> resolver_item = LazyOptional.of(() -> patterns.handler(
		((slot, stack) -> stack.getCapability(MatteryCapability.PATTERN).isPresent())
	));

	private MatterGrid grid;

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("patterns", patterns.serializeNBT());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);
		patterns.deserializeNBT(nbt.get("patterns"));
	}

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			MatterGrid.scheduleDiscoverNeighbours(this, getBlockPos(), p_155231_);
	}

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return null;
	}

	@Nullable
	@Override
	public IPatternStorage getPatternStorage() {
		return valid ? this : null;
	}

	private boolean valid = true;

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	private final LazyOptional<BlockEntityPatternStorage> resolver = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid) {
			if (cap == MatteryCapability.PATTERN || cap == MatteryCapability.MATTER_CELL)
				return resolver.cast();

			if (cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
				return resolver_item.cast();
		}

		return super.getCapability(cap, side);
	}

	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.pattern_storage");

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new PatternStorageMenu(containerID, inventory, this);
	}

	@Nonnull
	@Override
	public Collection<PatternState> getStoredPatterns() {
		ArrayList<PatternState> list = new ArrayList<>();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> list.addAll(capability.getStoredPatterns()));
		return ImmutableList.copyOf(list);
	}

	@Override
	public int getCapacity() {
		AtomicInteger stored = new AtomicInteger();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> stored.getAndAdd(capability.getCapacity()));
		return stored.get();
	}

	@Override
	public int getStored() {
		AtomicInteger stored = new AtomicInteger();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> stored.getAndAdd(capability.getStored()));
		return stored.get();
	}

	@Override
	public void setRemoved() {
		if (level != null && !level.isClientSide && grid != null)
			grid.untrack(this);

		super.setRemoved();
	}

	@Override
	public PatternInsertStatus insertPattern(PatternState pattern, boolean only_update, boolean simulate) {
		for (IPatternStorage storage : patterns.capabilityIterator(MatteryCapability.PATTERN)) {
			var status = storage.insertPattern(pattern, only_update, simulate);

			if (status.status() != PatternInsertStatus.Status.FAIL) {
				if (!simulate) {
					setChanged();

					if (grid != null) {
						if (status.status() == PatternInsertStatus.Status.INSERTED) {
							grid.onPatternAdded(status.new_state());
						} else {
							grid.onPatternUpdated(status.new_state(), status.old_state());
						}
					}
				}

				return status;
			}
		}

		return new PatternInsertStatus();
	}
}
